from .. import exceptions

import zope.interface
from zope.interface.verify import verifyObject
from zope.interface.exceptions import BrokenImplementation
from vagoth.interfaces.registry import IRegistry
from vagoth.interfaces.nodedoc import INodeDoc

class RegistryMixin(object):
    def mixin_setUp(self):
        # declare that we implement the IRegistry zope interface
        zope.interface.implementer(IRegistry)(self.registry.__class__)
        # add an initial test node
        self.registry.add_node("0xdeadbeef",
            node_type="hv",
            node_name="node001.example.com",
            tenant="mytenant",
            definition={ "name": "0xdeadbeef", "fqdn": "node001.example.com" },
            metadata={ "mymetakey": "mymetavalue" },
            tags={"tag1":True, "tag2":"somevalue"},
            unique_keys=["node001_uniquekey"])

    def test_initial_object(self):
        node = self.registry.get_node("0xdeadbeef")
        self.assertEqual("0xdeadbeef", node.id)
        self.assertEqual("node001.example.com", node.name)
        self.assertEqual("mytenant", node.tenant)
        self.assertEqual(["node001_uniquekey"], node.unique_keys)
        self.assertEqual(node.definition,
            { "name": "0xdeadbeef", "fqdn": "node001.example.com" })
        self.assertEqual(node.metadata,
            { "mymetakey": "mymetavalue" })

    def test_get_nodes(self):
        self.assertEqual(1, len(list(self.registry.get_nodes())))
        self.assertEqual(1,
            len(list(self.registry.get_nodes(tenant="mytenant"))))
        self.assertEqual(0,
            len(list(self.registry.get_nodes(tenant="nonexistant"))))
        self.assertEqual(1,
            len(list(self.registry.get_nodes(tags={"tag1":None}))))
        self.assertEqual(1,
            len(list(self.registry.get_nodes(tags={"tag2":"somevalue"}))))
        self.assertEqual(0,
            len(list(self.registry.get_nodes(tags={"tag2":"wrongvalue"}))))
        self.assertEqual(1,
            len(list(self.registry.get_nodes(parent=None))))
        self.assertEqual(1,
            len(list(self.registry.get_nodes(node_type="hv"))))
        self.assertEqual(1,
            len(list(self.registry.get_nodes(
                tenant="mytenant",
                parent=None,
                tags={"tag2":"somevalue"},
                node_type="hv"))))
        self.assertEqual(0,
            len(list(self.registry.get_nodes(
                tenant="wrongtenant",
                parent=None,
                tags={"tag2":"somevalue"},
                node_type="hv"))))

    def test_blobs(self):
        mydict = {"mykey": "myvalue"}
        self.registry.set_blob("0xdeadbeef", "blob_one", "one")
        self.registry.set_blob("0xdeadbeef", "blob_two", 2)
        self.registry.set_blob("0xdeadbeef", "blob_dict", mydict)
        self.assertEqual(self.registry.get_blob("0xdeadbeef", "blob_one"), "one")
        self.assertEqual(self.registry.get_blob("0xdeadbeef", "blob_two"), 2)
        self.assertEqual(self.registry.get_blob("0xdeadbeef", "blob_dict"), mydict)
        self.assertEqual(self.registry.get_blob("0xdeadbeef", "blob_unset"), None)

    def test_get_nodes_with_parent(self):
        self.registry.add_node("othernode", node_name="othernode", node_type="hv", tenant=None)
        self.registry.set_parent('0xdeadbeef', 'othernode')
        nodes = list(self.registry.get_nodes_with_parent("othernode"))
        self.assertEqual(len(nodes), 1)
        self.assertEqual(nodes[0].id, '0xdeadbeef')

    def test_get_node_by_name(self):
        node = self.registry.get_node_by_name("node001.example.com")
        self.assertEqual(node.id, '0xdeadbeef')

    def test_get_node_by_key(self):
        node = self.registry.get_node_by_key("node001_uniquekey")
        self.assertEqual(node.id, '0xdeadbeef')

    def test_get_nodes_with_tags_existence(self):
        nodes = list(self.registry.get_nodes_with_tags({"tag1": None}))
        self.assertEqual(len(nodes), 1)
        self.assertEqual(nodes[0].id, '0xdeadbeef')

    def test_get_nodes_with_tags_value(self):
        nodes = list(self.registry.get_nodes_with_tags({"tag2": "somevalue"}))
        self.assertEqual(len(nodes), 1)
        self.assertEqual(nodes[0].id, '0xdeadbeef')

    def test_get_nodes_with_type(self):
        nodes = list(self.registry.get_nodes_with_type("hv"))
        self.assertEqual(len(nodes), 1)
        self.assertEqual(nodes[0].id, '0xdeadbeef')

    def test_set_parent(self):
        self.registry.add_node("othernode", node_name="othernode", node_type="hv", tenant=None)
        self.registry.set_parent('0xdeadbeef', 'othernode')
        self.assertEqual(self.registry.nodes['0xdeadbeef']['parent'], 'othernode')

    def test_list_nodes(self):
        nodes = self.registry.list_nodes()
        self.assertEqual(nodes, ['0xdeadbeef'])

    def test_change_tags(self):
        self.registry.set_node("0xdeadbeef", tags={"xyzzy":"xyzzy"})
        self.assertIn("xyzzy", self.registry.get_node('0xdeadbeef').tags)

    def test_update_metadata(self):
        # add new keys
        self.registry.update_metadata("0xdeadbeef", {
            "one": "two"
        })
        metadata = self.registry.get_node('0xdeadbeef').metadata
        self.assertEqual(metadata['one'], 'two')
        # ensure initial metadata is still present
        self.assertEqual(metadata['mymetakey'], 'mymetavalue')
        # ensure it doesn't overwrite existing keys
        self.registry.update_metadata("0xdeadbeef", {"three": "four"})
        metadata = self.registry.get_node('0xdeadbeef').metadata
        self.assertEqual(metadata['one'], 'two')
        # ensure it deletes keys..
        self.registry.update_metadata("0xdeadbeef", {}, ["one"])
        metadata = self.registry.get_node('0xdeadbeef').metadata
        self.assertNotIn("one", metadata)
        # ..but not all keys
        self.assertIn("three", metadata)

    def test_change_name(self):
        self.registry.set_node("0xdeadbeef", node_name="foo.example.com")
        node = self.registry.get_node("0xdeadbeef")
        self.assertEqual(node.name, "foo.example.com")
        self.assertEqual("0xdeadbeef",
            self.registry.get_node_by_key("VAGOTH_NAME_foo.example.com").id)
        self.assertRaises(exceptions.NodeNotFoundException,
            self.registry.get_node_by_key, "VAGOTH_NAME_node001.example.com")

    def test_change_keys(self):
        self.registry.set_node("0xdeadbeef", unique_keys=["MASTER"])
        node = self.registry.get_node_by_key("MASTER")
        self.assertIn("MASTER", node.unique_keys)
        self.assertEqual(node.id, "0xdeadbeef")
        self.assertRaises(exceptions.NodeNotFoundException,
            self.registry.get_node_by_key, "node001_uniquekey")

    def test_delete_node(self):
        self.registry.delete_node("0xdeadbeef")
        self.assertRaises(exceptions.NodeNotFoundException,
            self.registry.get_node, "0xdeadbeef")

    def test_contains(self):
        self.assertIn("0xdeadbeef", self.registry)

    def test_interface(self):
        verifyObject(IRegistry, self.registry)



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

import zope.interface as ZI

class IRegistry(ZI.Interface):
    """
    The registry provides the data store for Vagoth.  It maintains
    nodes, their type, definition, metadata, tags, unique keys,
    blobs, and whether they have a parent.  It lets you add new nodes,
    change node attributes, and atomically set the parent and
    update metadata.
    """
    def __init__(manager, config):
        """
        Registry is initialized with the manager, and its config dict
        """

    def list_nodes():
        """
        Return an iterable of node_id's in the registry
        """

    def get_node(node_id):
        """Return the node identified by node_id

        :param node_id: node identifier
        :returns: INodeDoc
        """

    def get_node_by_name(node_name):
        """Return an the node with the name of node_name

        :param node_name: human-friendly (but still unique) name of node
        :returns: INodeDoc
        """

    def get_node_by_key(unique_key):
        """Return the node with the given unique key

        :param unique_key: unique key to search for
        :returns: INodeDoc
        """

    def get_nodes(tenant=None, node_type=None, tags=None, parent=None):
        """
        Returns all nodes which match the supplied filters.

        :returns: Iterable of INodeDoc's
        """

    def get_nodes_with_type(node_type):
        """Return all nodes with a matching type

        :param node_type: node type to match
        :returns: Iteratable of INodeDoc's
        """

    def get_nodes_with_tags(tag_matches):
        """Returns all nodes which match the given tags

        :param tag_matches: dict of key-value pairs that must match
        :returns: Iterable of INodeDoc's
        """

    def get_nodes_with_parent(node_parent):
        """Returns all node who have the given parent

        :param node_parent: parent's node id
        :returns: Iterable of INodeDoc's
        """

    def set_parent(node_id, parent_node_id):
        """
        Set the child's parent to parent_node_id.

        If parent_node_id is None, this will remove the child-parent relationship.

        If a parent is already assigned, this will throw an exception.
        """

    def add_node(node_id, node_name, node_type, tenant, definition=None, metadata=None, tags=None, unique_keys=None):
        """
        Add a node to the registry
        """

    def set_node(node_id, node_name=None, tenant=None, definition=None, metadata=None, tags=None, unique_keys=None):
        """
        Change attributes of an existing node in the registry.
        All parameters except node_id are optional.

        :param node_id: unique ID of the node
        :param node_name: unique name of the node
        :param tenant: name/identifier of the tenant (owner/project)
        :param definition: definition dictionary
        :param metadata: metadata dictionary
        :param tags: tags dictionary
        :param unique_keys: list of unique keys to claim
        """

    def update_metadata(node_id, extra_metadata, delete_keys=None):
        """
        Atomically call metadata.update(extra_metadata) and delete each key
        in delete_keys.

        This is available because set_node() will overwrite the existing
        metadata en-masse, which with multiple systems could lead to a race
        condition and unexpected results.

        Metadata, unlike definition, tags, and keys, is likely to be changed
        regularly by multiple processes, and it therefore justifies an extra
        method.

        :param node_id: unique ID of the node
        :param extra_metadata: dict of metadata to add to existing metadata dict
        :param delete_keys: list of metadata keys to delete
        """

    def set_blob(node_id, key, value):
        """
        Set a blob for the given node_id and (string) key to value.

        Blobs are undefined types, and may be stored separately to
        the rest of the node metadata in a given registry implementation.

        It's recommended to use 'metadata' instead, if it makes sense.
        """

    def get_blob(node_id, key):
        """
        Return the blob for the given node_id and (string) key.
        """

    def __contains__(node_id):
        """
        A shortcut to see if the given node is in the registry
        """



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

#
# A registry using a filesystem as the database.
#

import couchdb
from .. import exceptions
from .. import utils
from nodedoc import NodeDoc

class CouchRegistry(object):
    """
    self.nodes is set to the nodes table.
    self.unique is set to the unique keys table.

    Example node doc::

        "node001": {
            "node_id": "node001",
            "type": "type of node",
            "name": "nice node name",
            "tenant": "tenant identifier",
            "definition": { ... node definition ... },
            "metadata": { ... node metadata ... },
            "unique_keys": [ "one", "two", "three" ],
            "parent": None,
        }
    """
    def __init__(self, manager, config):
        self.manager = manager
        self.couchdb = couchdb.Server(config['couch_server'])
        self.nodes = self.couchdb[config['couch_nodes_table']]
        self.unique = self.couchdb[config['couch_unique_table']]

    def list_nodes(self):
        return list(self.nodes)

    def get_nodes(self, tenant=False, node_type=False, tags=False, parent=False):
        for node_id in self.nodes:
            node = self.nodes[node_id]
            if tenant is not False and node.get("tenant", None) != tenant:
                continue
            if parent is not False and not node["parent"] == parent:
                continue
            if node_type and node["type"] != node_type:
                continue
            if tags and not utils.matches_tags(tags, node["tags"]):
                continue
            yield NodeDoc(self, node)

    def __contains__(self, node_id):
        return node_id in self.nodes

    def get_node(self, node_id):
        "return dict for node"
        node = self.nodes.get(node_id, None)
        if node:
            return NodeDoc(self, node)
        raise exceptions.NodeNotFoundException("Node {0} not found in registry.".format(node_id))

    def get_node_by_name(self, node_name):
        "return dict for node"
        key = "VAGOTH_NAME_"+node_name
        if key in self.unique:
            node_id = self.unique[key]['node_id']
            if node_id in self.nodes:
                return NodeDoc(self, self.nodes[node_id])
        raise exceptions.NodeNotFoundException("Node with name {0} not found in registry.".format(node_name))

    def get_node_by_key(self, unique_key):
        "return the node with given unique key"
        if unique_key and unique_key in self.unique:
            node_id = self.unique[unique_key]['node_id']
            return NodeDoc(self, self.nodes.get(node_id, None))
        raise exceptions.NodeNotFoundException("Node with key {0} not found in registry.".format(unique_key))

    def get_nodes_with_type(self, node_type):
        "return list of nodes with type"
        return self.get_nodes(node_type=node_type)

    def get_nodes_with_tags(self, tag_matches):
        """return nodes with matching tags

        :param tag_matches: key/value pairs to match. If the value is None,
            check for key existence only.
        :returns: iterable of node dict's
        """
        return self.get_nodes(tags=tag_matches)

    def get_nodes_with_parent(self, parent=None):
        "return list of nodes with given parent"
        return self.get_nodes(parent=parent)

    def set_parent(self, node_id, parent_node_id):
        """Set parent_node_id atomically, and don't overwrite another"""
        doc = self.nodes[node_id]
        current_parent = doc.get('parent', None)
        if current_parent and parent_node_id is not None:
            raise exceptions.NodeAlreadyHasParentException("Node {0} already has a parent of {1}".format(node_id, current_parent))
        doc['parent'] = parent_node_id
        try:
            self.nodes.save(doc)
        except couchdb.http.ResourceConflict as e:
            raise exceptions.RegistryException(*e.args)

    def _claim_unique_keys(self, node_id, new_keys, old_keys=None):
        """
        Make a claim for node_id on new_keys, and release any old keys.

        We want to do this in a safe fashion.. so we should claim
        all new keys first, then release any old keys, and clean up
        in the event of any conflict.
        """
        assert node_id is not None
        claimed_keys = []
        for key in new_keys:
            doc = self.unique.get(key, None)
            if doc:
                if doc['node_id'] == node_id:
                    continue
                else:
                    raise exceptions.UniqueConstraintViolation("Key {0} is not unique.".format(key))
            try:
                self.unique[key] = { "node_id": node_id }
                claimed_keys.append(key)
            except couchdb.http.ResourceConflict:
                for delkey in claimed_keys:
                    del self.unique[delkey]
        if old_keys:
            for old_key in old_keys:
                if old_key in new_keys:
                    continue
                try:
                    doc = self.unique[old_key]
                except couchdb.http.ResourceNotFound:
                    continue
                if doc.get('node_id', None) == node_id:
                    try:
                        self.unique.delete(doc)
                    except couchdb.http.ResourceConflict:
                        raise # FIXME
                        # not good - let a cleanup job fix it later
                        pass
                # else: DB is out of sync - let a cleanup job fix it later
        #
        # At this point, all new keys have been claimed, and any old keys
        # should have been released.  Even if some old keys remain, our
        # claim on new keys is assured (as much as it can be without rogue
        # processes)
        #

    def add_node(self, node_id, node_name, node_type, tenant=None, definition=None, metadata=None, unique_keys=None, tags=None):
        """
        Create new node in registry
        """
        if unique_keys is None: unique_keys = []
        if metadata is None: metadata = {}
        if tags is None: tags = []
        # create a bare document as a reservation
        try:
            self.nodes[node_id] = {
                "node_id": node_id,
                "type": node_type, # only set once
                "name": None,
                "definition": {},
                "metadata": {},
                "unique_keys": [],
                "tags": [],
                "parent": None,
                "tenant": tenant,
            }
        except couchdb.http.ResourceConflict:
            raise exceptions.NodeAlreadyExistsException("Node {0} already exists in registry".format(node_id))
        doc = self.nodes[node_id]
        # add node_name to unique keys
        node_name_key = "VAGOTH_NAME_"+node_name
        if node_name_key not in unique_keys:
            unique_keys.insert(0, node_name_key)
        # claim all unique keys, or abort
        try:
            self._claim_unique_keys(node_id, unique_keys)
        except exceptions.UniqueConstraintViolation:
            self.nodes.delete(doc)
            raise
        # all unique keys have been claimed, so populate values
        doc['definition'] = definition
        doc['metadata'] = metadata
        doc['name'] = node_name
        doc['tags'] = tags
        doc['unique_keys'] = unique_keys[1:] # treat name specially
        self.nodes.save(doc)

    def _force_node_save(self, doc, count=5):
        """
        Attempt to save a node, despite any ResourceConflict's.
        """
        while count > 0:
            count -= 1
            try:
                self.nodes.save(doc)
                return
            except couchdb.http.ResourceConflict:
                newdoc = self.nodes[doc['_id']]
                doc['_rev'] = newdoc['_rev']
        raise exceptions.RegistryException("Could not write node %s to DB" % (doc['_id']))

    def set_node(self, node_id, node_name=None, tenant=None, definition=None, metadata=None, unique_keys=None, tags=None, **blobs):
        """Change an existing node definition"""
        try:
            doc = self.nodes[node_id]
        except couchdb.http.ResourceNotFound:
            raise exceptions.NodeNotFoundException("Node {0} not found in registry.".format(node_id))
        new_name_key = None
        old_name_key = None
        if definition:
            doc['definition'] = definition
        if metadata:
            doc['metadata'] = metadata
        if tags:
            doc['tags'] = tags
        if tenant:
            doc['tenant'] = tenant
        # update name (remember to cleanup old_name_key if set,
        #               or new_name_key on failure)
        if node_name and doc['name'] != node_name:
            new_name_key = 'VAGOTH_NAME_'+node_name
            self._claim_unique_keys(node_id, [new_name_key])
            old_name_key = 'VAGOTH_NAME_'+doc['name']
            doc['name'] = node_name
        if unique_keys:
            try:
                self._claim_unique_keys(node_id, unique_keys, doc['unique_keys'])
                doc['unique_keys'] = unique_keys
            except exceptions.UniqueConstraintViolation:
                if new_name_key:
                    self._claim_unique_keys(node_id, [], [new_name_key])
                raise
        if blobs:
            if "blobs" not in doc:
                doc["blobs"] = {}
            for k,v in blobs.items():
                if v is None:
                    if k in doc["blobs"]:
                        del doc["blobs"][k]
                else:
                    doc["blobs"][k] = v
            doc["blobs"] = blobs
        self._force_node_save(doc)
        if old_name_key:
            del self.unique[old_name_key]

    def update_metadata(self, node_id, extra_metadata=None, delete_keys=None):
        """Update metadata with extra_metadata, and delete any keys in delete_keys"""
        doc = self.nodes[node_id]
        if extra_metadata:
            doc['metadata'].update(extra_metadata)
        if delete_keys:
            for key in delete_keys:
                if key in doc['metadata']:
                    del doc['metadata'][key]
        try:
            self.nodes.save(doc)
        except couchdb.http.ResourceConflict as e:
            raise exceptions.RegistryException(*e.args)

    def set_blob(self, node_id, key, value):
        """Set a blob for the given node_id and key to the given value

        In this implementation, we store blobs as part of the node doc.
        """
        try:
            doc = self.nodes[node_id]
        except couchdb.http.ResourceNotFound:
            raise exceptions.NodeNotFoundException("Node {0} not found in registry.".format(node_id))
        if not doc.has_key("blobs"):
            doc["blobs"] = {}
        if value is None and key in doc["blobs"]:
            del doc["blobs"][key]
        else:
            doc["blobs"][key] = value
        try:
            self.nodes.save(doc)
        except couchdb.http.ResourceConflict as e:
            raise exceptions.RegistryException(*e.args)

    def get_blob(self, node_id, key):
        """Return the blob for the given node_id and key, or None

        In this implementation, we store blobs as part of the node doc.
        """
        node = self.nodes.get(node_id, None)
        if not node:
            raise exceptions.NodeNotFoundException("Node {0} not found in registry.".format(node_id))
        blobs = node.get("blobs", {})
        return blobs.get(key, None)

    def delete_node(self, node_id):
        """Delete the node identified by node_id"""
        try:
            doc = self.nodes[node_id]
        except couchdb.http.ResourceNotFound:
            raise exceptions.NodeNotFoundException("Node {0} not found in registry".format(node_id))
        unique_keys = doc.get('unique_keys', [])
        name_key = 'VAGOTH_NAME_'+doc['name']
        unique_keys.insert(0, name_key)
        self._claim_unique_keys(node_id, [], unique_keys)
        del self.nodes[node_id]



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

"""
Test vagoth.registry.couch_registry.CouchRegistry
"""

import unittest
from ..registry.couch_registry import CouchRegistry
import uuid
import couchdb
from .. import exceptions
from registry_mixin import RegistryMixin

rev_counter = 0

NO_DEFAULT=uuid.uuid4()

# an in-memory couchdb table mocker
class CouchTableMock:
    def __init__(self):
        self.data = {}
    def __setitem__(self, k, v):
        global rev_counter
        assert type(v) == dict
        if "_id" in v:
            if v["_id"] != k:
                raise couchdb.http.ResourceConflict("differing id in dict value")
        else:
            v["_id"] = k
        if k in self.data:
            if self.data[k]["_id"] != k:
                raise couchdb.http.ResourceConflict("differing id")
            if "_rev" in self.data[k] and v.get("_rev", None) != self.data[k]["_rev"]:
                raise couchdb.http.ResourceConflict("stale rev")
        if "_rev" not in v:
            v["_rev"] = rev_counter
            rev_counter += 1
        self.data[k] = v
    def __iter__(self):
        return iter(self.data)
    def __getitem__(self, k):
        return self.data[k]
    def __delitem__(self, k):
        del self.data[k]
    def get(self, k, default=NO_DEFAULT):
        if default == NO_DEFAULT:
            return self.data.get(k)
        else:
            return self.data.get(k, default)
    def __contains__(self, k):
        return k in self.data
    def save(self, doc):
        if "_id" in doc:
            self[doc["_id"]] = doc
    def delete(self, doc):
        if "_id" in doc and "_rev" in doc:
            _id, _rev = doc["_id"], doc["_rev"]
            if _id in self.data:
                if self.data[_id]["_rev"] == _rev:
                    del self.data[_id]
                else:
                    raise couchdb.http.ResourceConflict("stale rev")
            else:
                raise KeyError("not found in table")
        else:
            raise Exception("trying to delete a doc w/o _id and _rev: %r" % (doc,))

class CouchServer:
    def __init__(self, server):
        self.server = server
    def __getitem__(self, table_name):
        return CouchTableMock()

class test_CouchRegistry(unittest.TestCase, RegistryMixin):
    def setUp(self):
        self.tmp_Server = couchdb.Server
        couchdb.Server = CouchServer
        self.registry = CouchRegistry(None, {
            "couch_server": "http://example.com:5934",
            "couch_nodes_table": "vagoth/nodes",
            "couch_unique_table": "vagoth/nodes",
        })
        self.mixin_setUp()

    def tearDown(self):
        # restore sanity to the world
        couchdb.Server = self.tmp_Server

    def test_couch_claim_unique_keys(self):
        self.registry._claim_unique_keys("0x1", ["NEW"])
        self.assertEqual(len(self.registry.unique.data), 3)
        self.assertIn("NEW", self.registry.unique.data)
        self.assertTrue(self.registry.unique.data["NEW"]["node_id"] == "0x1")
        self.registry._claim_unique_keys("0x1", ["NEWER"], ["NEW"])
        self.assertTrue(len(self.registry.unique.data) == 3)
        self.assertNotIn("NEW", self.registry.unique.data)
        self.assertIn("NEWER", self.registry.unique.data)
        self.assertEqual(self.registry.unique.data["NEWER"]["node_id"], "0x1")

    def test_couch_claim_unique_keys_with_conflict(self):
        self.registry._claim_unique_keys("0x1", ["NEW"])
        self.assertRaises(exceptions.UniqueConstraintViolation,
            self.registry._claim_unique_keys, "0x2", ["NEW"])
        self.assertEqual(self.registry.unique.data["NEW"]["node_id"], "0x1")

    def test_couch_couch_initial_object(self):
        self.assertEqual(len(self.registry.nodes.data), 1)
        self.assertTrue("0xdeadbeef" in self.registry.nodes.data)
        self.assertEqual(len(self.registry.unique.data), 2)
        self.assertIn("VAGOTH_NAME_node001.example.com", self.registry.unique)
        self.assertIn("node001_uniquekey", self.registry.unique)



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

"""
This scheduler is suitable for small installations and frontend tools where
making synchronous calls to the hypervisor driver is appropriate.

For example, a command line VM management tool.
"""

from .. import exceptions

class SyncJobScheduler(object):
    """
    A synchronous job "scheduler"
    """
    def __init__(self, manager, config):
        self.manager = manager
        self.config = config

    def action(self, queue_name, action, **kwargs):
        """
        This calls self.manager.action() to run the action
        in the current thread.
        """
        self.manager.action(action, **kwargs)

    def cleanup(self):
        pass



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

#
# Registry of nodes
#

import pickle
import os.path
from dict_registry import DictRegistry
import fcntl

class LockFile(object):
    """
    A simple lock file that can be used in a "with x" statement.
    """
    def __init__(self, lockfile):
        self.lockfile = lockfile
        self.locked = False
        self.loaded = False
    def __enter__(self):
        self.fd = open(self.lockfile, 'w')
        fcntl.lockf(self.fd, fcntl.LOCK_EX)
        self.locked = True
    def __exit__(self, *args):
        self.locked = False
        self.loaded = False
        fcntl.lockf(self.fd, fcntl.LOCK_UN)
        self.fd.close()

class PickleRegistry(DictRegistry):
    """
    A simple registry based on DictRegistry but saving the contents
    to a python pickle.  It uses a lockfile to ensure process-safety.

    Config variable `lockfile` will set the location of the lockfile.
    Config variable `filename` will set the location of the pickle file.
    """

    def __init__(self, manager, config):
        lock = LockFile(config.get("lockfile", "/var/lock/vagoth.pickledb.lock"))
        DictRegistry.__init__(self, manager, config, lock=lock)
        # replace lock with our own file-based lock

    def _load(self):
        if self.lock.locked and self.lock.loaded:
            return # locked and loaded, Sir!
        self.lock.loaded = True
        filename = self.config.get('filename', None)
        if filename and os.path.exists(filename):
            with open(filename,'rw') as fd:
                self.data = pickle.load(fd)
        else:
            self.data = {"vms":{},"nodes":{}}
        self.vms = self.data["vms"]
        self.nodes = self.data["nodes"]

    def _save(self):
        filename = self.config.get('filename', None)
        with open(filename+".new", 'w') as fd:
            pickle.dump(self.data, fd)
        os.rename(filename+".new", filename)



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

from ..exceptions import NodeAlreadyExistsException
from ..exceptions import NodeStillUsedException

class DummyProvisioner(object):
    """
    DummyProvisioner does the bare minimum for a provisioner,
    creating a node in the registry with the same
    information that was passed into it.
    """

    def __init__(self, manager, config):
        self.manager = manager

    def provision(self, node_id,
                  node_name=None,
                  node_type=None,
                  tenant=None,
                  definition=None,
                  metadata=None,
                  tags=None,
                  unique_keys=None):
        # ensure it's a new node
        if node_id in self.manager.registry:
            raise NodeAlreadyExistsException("Node {0} already exists in registry".format(node_id))
        self.manager.registry.add_node(node_id,
            node_name = node_name or node_id,
            node_type = node_type,
            tenant = tenant,
            definition = definition,
            metadata = metadata,
            tags = tags,
            unique_keys = unique_keys
        )

    def deprovision(self, node_id):
        node = self.manager.registry.get_node(node_id) # throws NodeNotFoundException
        parent_node_id = node['parent']
        # check for a parent
        if parent_node_id and parent_node_id in self.manager.registry:
            raise NodeStillUsedException("Node {0} has a parent assigned.".format(node_id))
        # check for children
        children = list(self.manager.registry.get_nodes_with_parent(node_id))
        if len(children) > 0:
            raise NodeStillUsedException("Node {0} has children assigned.".format(node_id))
        # node seems like an orphan, so it's ok to delete it
        self.manager.registry.delete_node(node_id)



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

from ..exceptions import DriverException
import subprocess
import json
import os

def local_call(action, **kwargs):
    p = subprocess.Popen(["geats_jsonagent"], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    request = json.dumps({"action": action, "data": kwargs})
    p.stdin.write(request)
    p.stdin.close()
    result = p.stdout.read()
    exit_code = p.wait()
    if exit_code == 0:
        jresult = json.loads(result)
        assert type(jresult) == dict
        if jresult.get("success", False):
            return jresult.get("data", None)
        else:
            errorcode = jresult.get("errorcode", "undefined")
            message = jresult.get("message", "check logs")
            exceptionmsg = "%s: %s" % (errorcode, message)
            raise DriverException(exceptionmsg)
    else:
        raise DriverException("geats_jsonagent exited with %d" % (exit_code,))




class GeatsLocal(object):
    """
    Driver to talk to a local Geats install.
    """
    def __init__(self, manager, local_config):
        self.config = local_config

    def _call(self, action, node=None, timeout=60, **kwargs):
        return local_call(action, **kwargs)

    def _call_single(self, action, node, vm, timeout=60, **kwargs):
        vm_name = vm.node_id
        return local_call(action, vm_name=vm.node_id, **kwargs)
    _call_single_exc = _call_single

    def _call_boolean(self, action, node, vm, timeout=60, **kwargs):
        result = self._call_single(action, node, vm, timeout, **kwargs)
        return True

    def provision(self, node, vm):
        """Request a node to define & provision a VM"""
        return self._call_single_exc("provision", node, vm, definition=vm.definition)

    def define(self, node, vm):
        """Request node to define a VM"""
        return self._call_single_exc("define", node, vm, definition=vm.definition)

    def undefine(self, node, vm):
        """Request node to undefine a VM"""
        return self._call_boolean("undefine", node, vm)

    def deprovision(self, node, vm):
        """Request node to undefine & deprovision a VM"""
        return self._call_boolean("deprovision", node, vm)

    def start(self, node, vm):
        """Request node to start the VM"""
        return self._call_boolean("start", node, vm, timeout=10)

    def reboot(self, node, vm):
        """Request node to reboot the VM"""
        return self._call_boolean("reboot", node, vm, timeout=10)

    def stop(self, node, vm):
        """Request node to stop (forcefully) the VM"""
        return self._call_boolean("stop", node, vm, timeout=10)

    def shutdown(self, node, vm):
        """Request node to shutdown (nicely) the VM"""
        return self._call_boolean("shutdown", node, vm, timeout=10)

    def info(self, node, vm):
        """Request information about the given VM from the node"""
        info = self._call_single_exc("info", node, vm, timeout=5)
        if info:
            info[u"node"] = unicode(node.node_id)
        return info

    def status(self, node=None):
        """Request information about all VMs from the node"""
        res = self._call("status", node, timeout=5)
        if len(res) == 0:
            raise DriverException("No results received for %s" % (node))
        nodes = []
        node_name = os.uname()[1]
        if node is None:
            pass # just us, then
        elif node.node_id != node_name:
            # we only return information about the current node..
            raise DriverException("Cannot contact remote node '%s' using GeatsLocal driver" % (node.node_id,))
        for vm in res["vms"].values():
                vm[u"_parent"] = node_name
                vm[u"_name"] = vm["definition"]["name"]
                vm[u"_type"] = "vm"
                nodes.append(vm)
        # add hypervisor server
        hv = res.get("status", {})
        hv[u"_parent"] = None
        hv[u"_name"] = node_name
        hv[u"_type"] = "hv"
        nodes.append(hv)
        return nodes

    def migrate(self, node, vm, destination_node):
        """Request the node to migrate the given VM to the destination node"""
        return NotImplemented



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

from config import Config
import exceptions
import logging

class Manager(object):
    """
    The Manager class is the front door to the Vagoth API.
    Use manager.get_manager() and it will be cached and returned multiple times.

    You can pass in a different config object, otherwise it will instantiate
    vagoth.config_.Config
    """
    def __init__(self, config=None):
        config = self.config = config or Config()
        # self.registry
        registry_factory, registry_config = config.get_factory("registry")
        self.registry = registry_factory(self, registry_config)

        # self.scheduler
        sched_factory, sched_config = config.get_factory("scheduler")
        self.scheduler = sched_factory(self, sched_config)

        # self.provisioner
        provisioner_factory, provisioner_config = self.config.get_factory("provisioner")
        self.provisioner = provisioner_factory(self, provisioner_config)

    def _instantiate_node(self, nodedoc):
        if nodedoc:
            node_factory = self.config.get_node_factory(nodedoc.type)
            if node_factory:
                return node_factory(self, nodedoc)
            else:
                raise exceptions.UnknownNodeType("Unknown node type: %r" % (nodedoc.type))

    def get_node(self, node_id):
        """Return the node instance for this node_id"""
        nodedoc = self.registry.get_node(node_id) # can throw exception
        return self._instantiate_node(nodedoc)

    def get_nodes(self, tenant=False, node_type=False, tags=False, parent=False):
        """Return an iterable of all node instances"""
        for nodedoc in self.registry.get_nodes(tenant=tenant, node_type=node_type, tags=tags, parent=parent):
            yield self._instantiate_node(nodedoc)

    def get_node_by_name(self, node_name):
        """Return the node instance for the given node_name"""
        nodedoc = self.registry.get_node_by_name(node_name) # can throw exception
        return self._instantiate_node(nodedoc)

    def get_node_by_key(self, node_key):
        """Return the node instance which has the given unique key"""
        nodedoc = self.registry.get_node_by_key(node_key)
        return self._instantiate_node(nodedoc)

    def list_nodes(self):
        """Return an iterable of all the node_id in the registry"""
        return self.registry.list_nodes()

    def action(self, action, **kwargs):
        """Call the given action with the given arguments"""
        action_func = self.config.get_action(action)
        if action_func:
            try:
                action_func(self, **kwargs)
            except:
                logging.debug("Exception while executing action %s" % (action,), exc_info=True)
                raise

    def cleanup(self):
        """This must be called at shutdown time"""
        self.scheduler.cleanup()

manager = None
def get_manager(config=None):
    """
    Instantiate a new Manager instance and cache it,
    or return the cached instance.
    """
    global manager
    if manager is None:
        manager = Manager(config=config)
    return manager




#!/usr/bin/python
#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

from subprocess import Popen, PIPE
from os.path import abspath, dirname, join
import json

class MCollectiveException(Exception): pass


def mcollective_call(agent, action, identity=None, timeout=None, **kwargs):
    mcdict = {
        "agent": agent,
        "action": action,
        "arguments": kwargs,
    }
    if identity is not None:
        mcdict["identity"] = identity
    if timeout is not None:
        mcdict["timeout"] = timeout
    mcjson = json.dumps(mcdict)
    ruby_script=join(abspath(dirname(__file__)), "mc_json_rpc.rb")
    process = Popen([ruby_script, "-"], stdin=PIPE, stdout=PIPE)
    process.stdin.write(mcjson)
    process.stdin.close()
    result = process.stdout.read()
    process.stdout.close()
    process.wait()
    if process.returncode == 0:
        return json.loads(result)
    else:
        raise MCollectiveException(
            "mc-json-rpc.rb exited with {0}: {1}".format(
                process.returncode, result))



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

"""
Test vagoth.registry.couch_registry.CouchRegistry
"""

import unittest
from ..registry.dict_registry import DictRegistry
import uuid
import couchdb
from .. import exceptions
from registry_mixin import RegistryMixin

NO_DEFAULT=uuid.uuid4()

class testDictRegistry(unittest.TestCase, RegistryMixin):
    def setUp(self):
        self.registry = DictRegistry(None, {})
        self.mixin_setUp()

    def test_dict_initial_object(self):
        self.assertEqual(len(self.registry.nodes), 1)
        self.assertTrue("0xdeadbeef" in self.registry.nodes)
        self.assertEqual(len(self.registry.unique), 2)
        self.assertIn("VAGOTH_NAME_node001.example.com", self.registry.unique)
        self.assertIn("node001_uniquekey", self.registry.unique)



# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

#
# Some utility functions
#

def matches_tags(tag_matches, tags):
    """Do the given tag_matches match the given tags?

    If the tag_matches value is None, it only checks for tag existence.

    If the tag_matches value is not None, it does a direct comparison.
    For each key/value pair, check if the tag exists, and if the value is
    not None, if the value matches.

    :param tag_matches: key-value pairs we want to check for
    :param tags: key-value pairs that we'll check against
    :returns: bool
    """
    assert tag_matches
    assert type(tags) == dict

    for tag_name, tag_value in tag_matches.items():
        if tag_name not in tags:
            return False
        if tag_value is not None and tag_value != tags[tag_name]:
            return False
    return True




#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

from ..utils.mc_json_rpc import mcollective_call
from ..exceptions import DriverException

class GeatsMcollective(object):
    """
    Driver to talk to Geats using mcollective.  It uses the
    mcollective_call() function from utils.mc_json_rpc to
    launch ruby to make the call.
    """
    def __init__(self, manager, local_config):
        self.config = local_config

    def _call(self, action, node=None, timeout=60, **kwargs):
        if node:
            node_name = node.node_id
            return mcollective_call("geats", action, timeout=timeout, identity=node_name, **kwargs)
        else:
            return mcollective_call("geats", action, timeout=timeout, **kwargs)

    def _call_single(self, action, node, vm, timeout=60, **kwargs):
        if node is None:
            raise TypeError, "node argument is required"
        if vm is None:
            raise TypeError, "vm argument is required"
        node_name = node.node_id
        vm_name = vm.node_id
        responses = mcollective_call("geats", action, timeout=timeout, identity=node_name, vm_name=vm_name, **kwargs)
        # should only be one response
        for res in responses:
            return res["statuscode"], res["statusmsg"], res["data"]
        return None, None, None

    def _call_single_exc(self, action, node, vm, timeout=60, **kwargs):
        status, statusmsg, data = self._call_single(action, node, vm, timeout, **kwargs)
        if status is None:
            raise DriverException("No response received from %s" % (node,))
        if status != 0:
            raise DriverException(statusmsg) # FIXME
        return data

    def _call_boolean(self, action, node, vm, timeout=60, **kwargs):
        status, statusmsg, data = self._call_single(action, node, vm, timeout, **kwargs)
        if data:
            return True
        if status == 0:
            return True
        elif status is None:
            raise DriverException("No response received from %s" % (node,))
        else:
            raise DriverException(str(statusmsg))

    def provision(self, node, vm):
        """Request a node to define & provision a VM"""
        return self._call_single_exc("provision", node, vm, definition=vm.definition)

    def define(self, node, vm):
        """Request node to define a VM"""
        return self._call_single_exc("define", node, vm, definition=vm.definition)

    def undefine(self, node, vm):
        """Request node to undefine a VM"""
        return self._call_boolean("undefine", node, vm)

    def deprovision(self, node, vm):
        """Request node to undefine & deprovision a VM"""
        return self._call_boolean("deprovision", node, vm)

    def start(self, node, vm):
        """Request node to start the VM"""
        return self._call_boolean("start", node, vm, timeout=10)

    def reboot(self, node, vm):
        """Request node to reboot the VM"""
        return self._call_boolean("reboot", node, vm, timeout=10)

    def stop(self, node, vm):
        """Request node to stop (forcefully) the VM"""
        return self._call_boolean("stop", node, vm, timeout=10)

    def shutdown(self, node, vm):
        """Request node to shutdown (nicely) the VM"""
        return self._call_boolean("shutdown", node, vm, timeout=10)

    def info(self, node, vm):
        """Request information about the given VM from the node"""
        info = self._call_single_exc("info", node, vm, timeout=5)
        if info:
            info[u"node"] = unicode(node.node_id)
        return info

    def status(self, node=None):
        """Request information about all VMs from the node"""
        res = self._call("status", node, timeout=5)
        if len(res) == 0:
            raise DriverException("No results received for %s" % (node))
        nodes = []
        for node in res:
            if node["statuscode"] == 0:
                for vm in node["data"]["vms"].values():
                    vm[u"_parent"] = node["sender"]
                    vm[u"_name"] = vm["definition"]["name"]
                    vm[u"_type"] = "vm"
                    nodes.append(vm)
                # add hypervisor server
                hv = node["data"].get("status", {})
                hv[u"_parent"] = None
                hv[u"_name"] = node["sender"]
                hv[u"_type"] = "hv"
                nodes.append(hv)
        return nodes

    def migrate(self, node, vm, destination_node):
        """Request the node to migrate the given VM to the destination node"""
        return NotImplemented



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

import vagoth.exceptions
from exceptions import DriverException
import logging

class Monitor(object):
    """
    For each node, call driver.status, and update VM state.
    """
    def __init__(self, manager, config):
        self.manager = manager
        if config.get('create_missing', False) in ('true','yes',True):
            self.create_missing = True
        else:
            self.create_missing = False

    def create_vm(self, vm_name, vm_status, node):
        """
        Given the VM name, status, and the node it was found
        running on, call the provisioner to add it to the registry,
        then set the parent to the node.
        """
        provisioner = self.manager.provisioner
        logging.info("Creating VM %r on %r: %r" % (vm_name, node, vm_status))
        provisioner.provision(vm_name, node_name=vm_name, node_type='vm',
            definition = vm_status['definition'],
            metadata = { 'state': vm_name }
        )
        vm = self.manager.get_node(vm_name)
        self.manager.registry.set_parent(vm_name, node.node_id)
        return vm

    def unassign_vm(self, vm):
        """
        Unassign the given VM (set its parent to None)
        """
        self.manager.registry.set_parent(vm.node_id, None)
        vm.state = "unassigned"

    def update_node(self, node, status):
        """
        Given an iterable of VM statuses containing
        node definitions, check whether the registry
        is up to date, and if not, update it.

        If self.create_missing is set, create missing
        VMs.

        If a VM is not found on a node that it's
        assigned to, it will be unassigned from that
        node.
        """
        vms = {}
        for vm_status in status:
            vm_name = vm_status["_name"]
            if vm_name == node.node_id:
                # Update node's metadata based on node status
                # FIXME: a little hacky
                del vm_status["_name"]
                del vm_status["_type"]
                del vm_status["_parent"]
                self.manager.registry.update_metadata(node.node_id, vm_status)
                continue
            vms[vm_name] = vm_status
            vm_state, vm_target_state = vm_status["state"]
            try:
                vm = self.manager.get_node(vm_name)
            except vagoth.exceptions.NodeNotFoundException:
                if self.create_missing:
                    vm = self.create_vm(vm_name, vm_status, node)
                else:
                    logging.warn("VM {0} not found. Skipping.".format(vm_name))
                    continue

            # inconsistency detected:
            # if VM is running on a node, but isn't assigned to it, then assign it
            if vm and not vm.parent:
                self.manager.registry.set_parent(vm_name, node.node_id)
                logging.warn("Auto-assigning {0} to {1}".format(vm_name, node.node_id))

            # update state:
            if vm and vm.parent == node:
                if vm.state != vm_state:
                    vm.state = vm_state
                    logging.debug("Setting VM state for {0} to {1}".format(vm_name, vm_state))

        for vm in self.manager.get_nodes(parent=node.node_id):
            # any VMs assigned to node, but not active?
            if vm.node_id not in vms:
                logging.debug("Unassigning VM {0} from {1}".format(vm.node_id, node.node_id))
                self.unassign_vm(vm)

    def poll_nodes(self):
        """
        Poll each node individually (by calling driver.status(node)) and
        call self.update_node with the returned status.
        """
        for node in self.manager.get_nodes(node_type="hv"):
            driver = node.driver
            if driver:
                try:
                    node_status = driver.status(node)
                except DriverException as e:
                    logging.info(e.message)
                    continue
                self.update_node(node, node_status)



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

import utils

class Node(object):
    """
    A Node represents a node entry in the Registry.

    Node is an extensible wrapper for an INodeDoc instance, which is
    a read-only snapshot of a node from the registry.

    It's expected that specific node types will inherit this Node
    and add python-API-friendly functionality to it.
    """
    def __init__(self, manager, node_doc):
        """
        :param manager: vagoth.manager.Manager instance
        :param node_id: ID of node
        :param node_doc: must implement INodeDoc
        """
        self._manager = manager
        self._node_id = node_doc.id
        self._doc = node_doc

    def refresh(self):
        """
        Refresh node data by querying the registry for the latest INodeDoc
        """
        self._doc = self._manager.registry.get_node(self.node_id)

    # ensure it's read-only
    @property
    def node_id(self):
        """The unique and immutable id of this node"""
        return self._node_id

    @property
    def node_type(self):
        """The immutable type of this node"""
        return self._doc.type

    @property
    def name(self):
        """The unique but mutable name of this node"""
        return self._doc.name

    @property
    def definition(self):
        """The definition dictionary"""
        return self._doc.definition

    @property
    def metadata(self):
        """The metadata dictionary"""
        return self._doc.metadata

    @property
    def parent_id(self):
        """The node id of the parent node of this one, if set"""
        parent_id = self._doc.parent
        if parent_id:
            return parent_id

    @property
    def parent(self):
        """The parent node of this one, if set"""
        parent_id = self._doc.parent
        if parent_id:
            return self._manager.get_node(parent_id)

    @property
    def tags(self):
        """Dict of all tags for this node"""
        return self._doc.tags

    @property
    def tenant(self):
        return self._doc.tenant

    @property
    def unique_keys(self):
        """List of all unique keys for this node"""
        return list(self._doc.unique_keys)

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<Node %s at %x>" % (self.node_id, id(self))

    def __eq__(self, other):
        assert other is None or isinstance(other, Node)
        if other is None:
            return False
        return self.node_id == other.node_id

    def __ne__(self, other):
        return not self == other

    def matches_tags(self, tag_matches):
        """Does this node match the given tags?

        For each key/value pair, check if the tag exists, and if the value is
        not None, if the value matches.
        """
        return utils.matches_tags(tag_matches, self._doc.tags or {})



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

"""
Registry of Nodes, backed by dictionaries

It also calls _load and _save methods before hand after any
operations, even though it's probably a no-op here.  It uses
a lock (by default threading.RLock) to lock write operations.

A different lock can passed in as an extra argument.
"""

from .. import exceptions
from .. import utils
from nodedoc import NodeDoc
from threading import RLock

class DictRegistry(object):
    """
    A simple thread-safe (but not process-safe) registry.

    self.nodes is set to the nodes dictionary.

    Example node dict::

        self.nodes = { "node001": {
            "node_id": "node001",
            "name": "node001",
            "type": "vm",
            "definition": { ... node definition ... },
            "metadata": { ... node metadata ... },
            "tags": [ "tag1", "tag2" ],
            "unique_keys": [ "ip-1.2.3.4" ],
            "parent": None,
        } }
    """
    def __init__(self, manager, config, lock=None):
        self.manager = manager
        self.config = config
        self.lock = lock or RLock()
        self.nodes = {}
        self.unique = {}

    def _load(self):
        """Override to load the nodes and unique dicts"""
        pass

    def _save(self):
        """Override to save the nodes and unique dicts"""
        pass

    def __contains__(self, node_id):
        return node_id in self.nodes

    def list_nodes(self):
        """Return a list of all node_id"""
        return self.nodes.keys()

    def _get_node(self, node_id):
        """Return a node doc for the given node_id"""
        self._load()
        try:
            return self.nodes[node_id]
        except KeyError:
            raise exceptions.NodeNotFoundException("Node %s not found in registry" % (node_id,))

    def get_node(self, node_id):
        return NodeDoc(self, self._get_node(node_id))

    def get_node_by_name(self, node_name):
        """Return a node doc for the node with the given node_name"""
        self._load()
        for node in self.nodes.itervalues():
            if node.get('name', None) == node_name:
                return NodeDoc(self, node)
        raise exceptions.NodeNotFoundException("Node not found in registry with name: %s" % (node_name,))

    def get_node_by_key(self, key):
        """Return a node doc for the node with the given key"""
        self._load()
        node_id = self.unique.get(key, None)
        if node_id:
            return self.get_node(node_id)
        raise exceptions.NodeNotFoundException("Node not found in registry with key: %s" % (key,))

    def get_nodes(self, tenant=False, node_type=None, tags=None, parent=False):
        """Return an iterable of node docs"""
        self._load()
        for node in self.nodes.values():
            if tenant is not False and node["tenant"] != tenant:
                continue
            if node_type and node["type"] != node_type:
                continue
            if parent is not False and not node["parent"] == parent:
                continue
            if tags and not utils.matches_tags(tags, node["tags"]):
                continue
            yield NodeDoc(self, node)

    def get_nodes_with_type(self, node_type):
        """Return an iterable of node docs with the given type"""
        self._load()
        for node in self.nodes.itervalues():
            if node['type'] == node_type:
                yield NodeDoc(self, node)

    def get_nodes_with_tags(self, tag_matches):
        """Return an iterable of node docs with the given tag

        :param tag_matches: key/value pairs to match. If the value is None,
            check for key existence only.
        :returns: iterable of node dict's
        """
        self._load()
        for node in self.nodes.itervalues():
            tags = node.get('tags', {})
            if type(tags) == list:
                tags = dict([(x,True) for x in tags])
            for tag_name, tag_value in tag_matches.items():
                if tag_name not in tags:
                    continue
                if tag_value is not None and tag_value != tags[tag_name]:
                    continue
                yield NodeDoc(self, node)

    def get_nodes_with_parent(self, node_parent):
        """Return an iterable of node docs with the given parent id"""
        self._load()
        for node in self.nodes.itervalues():
            if node.get('parent', None) == node_parent:
                yield NodeDoc(self, node)

    def set_parent(self, node_id, parent_node_id):
        """
        Set the parent node id, but only if it's not already set.
        It can also be used to set the parent back to to None.
        """
        with self.lock:
            self._load()
            try:
                node = self.nodes[node_id]
            except KeyError:
                raise exceptions.NodeNotFoundException("Node not found: %s" % (node_id,))
            parent = node.get("parent", None)
            if parent_node_id is None:
                node["parent"] = None
            elif parent is None:
                if parent_node_id in self.nodes:
                    node['parent'] = parent_node_id
                else:
                    raise exceptions.NodeNotFoundException("Parent node not found: %s" % (parent_node_id,))
            else:
                raise exceptions.NodeAlreadyHasParentException("Node already has a parent. Unassign it first: %s" % (node_id,))
            self._save()

    def add_node(self, node_id, node_name, node_type, tenant, definition=None, metadata=None, tags=None, unique_keys=None):
        """
        Add a new node to the registry, ensuring that the node_id,
        node_name, and keys are unique.
        """
        with self.lock:
            self._load()
            if node_id in self.nodes:
                raise exceptions.NodeAlreadyExistsException("Node already exists in registry: %s" % (node_id,))
            node = {
                "node_id": node_id,
                "name": node_name,
                "tenant": tenant,
                "type": node_type,
                "definition": definition or {},
                "metadata": metadata or {},
                "tags": tags or [],
                "unique_keys": unique_keys or [],
                "parent": None,
            }
            namekey = "VAGOTH_NAME_%s" % node_name
            if namekey in self.unique:
                raise exceptions.UniqueConstraintViolation("Node name already taken: %s" % (node_name,))
            for key in (unique_keys or []):
                if key in self.unique:
                    raise exceptions.UniqueConstraintViolation("Unique key is already taken: %s" % (key,))
            self.unique[namekey] = node_id
            for key in (unique_keys or []):
                self.unique[key] = node_id
            self.nodes[node_id] = node
            self._save()

    def set_node(self, node_id, node_name=None, tenant=None, definition=None, metadata=None, tags=None, unique_keys=None):
        """
        Update the node specified by node_id, ensuring that all uniqueness constraints are
        still valid. No changes will be made if there is the chance of a name or unique key
        collision.
        """
        with self.lock:
            self._load()
            try:
                node = self.nodes[node_id]
            except KeyError:
                raise exceptions.NodeNotFoundException("Node not found: %s" % (node_id,))
            if node_name and node_name != node["name"]:
                namekey = "VAGOTH_NAME_%s" % (node_name,)
                if namekey in self.unique:
                    raise exceptions.UniqueConstraintViolation("Node name already taken: %s" % (node_name,))
            if unique_keys:
                for key in unique_keys:
                    if key in self.unique and self.unique[key] != node_id:
                        raise exceptions.UniqueConstraintViolation("Unique key is already taken: %s" % (key,))
            if tenant:
                node["tenant"] = tenant
            if definition:
                node["definition"] = definition
            if metadata:
                node["metadata"] = metadata
            if tags:
                node["tags"] = tags
            if unique_keys: # unique_keys are the new set of keys
                for key in node["unique_keys"]:
                    if key in self.unique and self.unique[key] == node_id and key not in unique_keys:
                        del self.unique[key]
                for key in unique_keys:
                    if key not in self.unique:
                        self.unique[key] = node_id
                node["unique_keys"] = unique_keys
            if node_name:
                oldnamekey = "VAGOTH_NAME_%s" % (node["name"],)
                newnamekey = "VAGOTH_NAME_%s" % (node_name,)
                if oldnamekey != newnamekey:
                    del self.unique[oldnamekey]
                    self.unique[newnamekey] = node_id
                node["name"] = node_name
            self._save()

    def set_blob(self, node_id, key, value):
        """Set a blob for the given node_id and key to the given value"""
        with self.lock:
            self._load()
            try:
                node = self.nodes[node_id]
            except KeyError:
                raise exceptions.NodeNotFoundException("Node not found: %s" % (node_id,))
            if not node.has_key("blobs"):
                node["blobs"] = {}
            if value is None and key in node["blobs"]:
                del node["blobs"][key]
            else:
                node["blobs"][key] = value
            self._save()

    def get_blob(self, node_id, key):
        """Return the blob for the given node_id and key, or None"""
        node = self._get_node(node_id)
        blobs = node.get("blobs", {})
        return blobs.get(key, None)

    def update_metadata(self, node_id, extra_metadata, delete_keys=None):
        """Atomically update the metadata dict, or delete keys from it"""
        with self.lock:
            self._load()
            try:
                node = self.nodes[node_id]
            except KeyError:
                raise exceptions.NodeNotFoundException("Node not found: %s" % (node_id,))
            metadata = node.get('metadata', {})
            for key in (delete_keys or []):
                if key in metadata:
                    del metadata[key]
            if extra_metadata:
                metadata.update(extra_metadata)
            self._save()

    def delete_node(self, node_id):
        """Delete the given node, freeing up its resources"""
        with self.lock:
            try:
                node = self.nodes[node_id]
            except KeyError:
                raise exceptions.NodeNotFoundException("Node not found: %s" % (node_id,))
            if node.get("parent", None):
                raise exceptions.NodeStillUsedException("Node still has a parent: %s" % (node_id,))
            children = list(self.get_nodes_with_parent(node_id))
            if children:
                raise exceptions.NodeStillUsedException("Node still has children: %s" % (node_id,))
            # not in use, so delete it and its unique keys
            namekey = "VAGOTH_NAME_%s" % (node["name"],)
            if namekey in self.unique and self.unique[namekey] == node_id:
                del self.unique[namekey]
            for key in node.get("unique_keys", []):
                if key in self.unique and self.unique[key] == node_id:
                    del self.unique[key]
            del self.nodes[node_id]
            self._save()



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

class DummyRegistry(object):
    def __init__(self):
        self.nodes = {}
        self.vms = {}

    def _get_vm_attr(vm_name, key):
        vm = self.vms.get(vm_name, None)
        if vm is None:
            return None
        return vm[key]

    def _set_vm_attr(vm_name, definition=None, metadata=None, state=None):
        vm = self.vms.get(vm_name, None)
        if vm is None:
            self.vms[vm_name] = {
                "definition": definition,
                "metadata": metadata,
                "state": state,
            }
            return
        if definition is not None:
            vm["definition"] = definition
        if metadata is not None:
            vm["metadata"] = metadata
        if state is not None:
            if not isinstance(state, basestring):
                raise TypeError("state must be a string value, not %s" % type(state))
            vm["state"] = state

    def get_vm_definition(self, vm_name):
        return self._get_vm_attr("definition")

    def set_vm_definition(self, vm_name, definition):
        self._set_vm_attr(vm_name, definition=definition)

    def get_vm_metadata(self, vm_name):
        return self._get_vm_attr("metadata")

    def set_vm_metadata(self, vm_name, metadata):
        self._set_vm_attr(vm_name, metadata=metadata)

    def get_vm_state(self, vm_name):
        return self._get_vm_attr("state")

    def set_vm_state(self, vm_name, state):
        return self._get_vm_attr("state", state=state)

    def define_vm(self, vm_name, definition, metadata=None, state='defined'):
        pass



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

import zope.interface as ZI

class IProvisioner(ZI.Interface):
    """
    The provisioner is called to add a node to the cluster, and to
    remove a node from the cluster.

    It is the responsibility of the Provisioner to ensure that a node
    is correct and valid before adding it to Vagoth.

    Depending on how you want to use Vagoth, you could simply pass through
    the definition as-is to the registry, or you may want to convert
    "vm_size": "small" into a fully fledged definition complete
    with IPs, storage, etc.

    You may also like to hook this into some DNS and DHCP API to create
    entries for the host.

    vagoth.exceptions.ProvisioningException is the only exception
    that should be handled.  All other exceptions are undefined
    behaviour.

    If ProvisioningException is raised during the provision() call
    then the node has not been provisioned in the cluster.

    If ProvisioningException is raised during the deprovision() call,
    then the node remains in the cluster.
    """
    def __init__(manager, config):
        """Takes the manager, local config dict and the global config object"""
    def provision(node_id, node_name=None, node_type=None, tenant=None, definition=None, metadata=None, tags=None, unique_keys=None):
        """
        Do all required steps to provision the given node in Vagoth, including adding to the registry.
        Throw a ProvisioningException if there were any issues.
        """
    def deprovision(node_id):
        """
        Do any cleanup of resources etc. before this VM definition is
        removed from the cluster.
        """



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

from threading import Thread
from Queue import Queue

class Worker(object):
    """
    Worker class for `ThreadScheduler`
    """
    def __init__(self, manager):
        self.manager = manager
        self.queue = Queue()

    def run(self):
        while True:
            action, kwargs = self.queue.get()
            if action is None:
                break
            try:
                self.manager.action(action, **kwargs)
            except Exception as e:
                print e # FIXME

    def stop(self):
        """Signal to the thread that it should quit"""
        self.queue.put( (None, None) )

    def action(self, action, **kwargs):
        """Add an action to the queue"""
        self.queue.put( (action, kwargs) )

class ThreadScheduler(object):
    """
    Runs a single worker in a Thread to process
    all actions in the background.
    """
    def __init__(self, manager, config):
        self.manager = manager
        self.config = config
        self.worker = Worker(manager)
        self.thread = Thread(target=self.worker.run)
        self.thread.start()

    def action(self, queue_name, action, **kwargs):
        """Schedule the specified action, ignoring queue_name"""
        self.worker.action(action, **kwargs)

    def cleanup(self):
        """Request the worker thread to exit"""
        self.worker.stop()



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

from ..node import Node

class Hypervisor(Node):
    """
    A basic Hypervisor class.  A hypervisor has a driver

    It inherits vagoth.Node_
    """

    # FIXME, duplicated in VirtualMachine (but not really a Node thing)
    @property
    def state(self):
        """Retrieve the state attribute from the metadata"""
        return self._doc.metadata.get('state', 'unknown')

    @state.setter
    def state(self, state):
        """Set the state attribute in the metadata"""
        self._doc.registry.update_metadata(self.node_id, { "state": state, })
        self.refresh()

    @property
    def children(self):
        """Return the children of this hypervisor"""
        return self._manager.get_nodes_with_parent(self._node_id)

    @property
    def driver_name(self):
        """Return the driver name for this hypervisor, or default"""
        driver_name = self.definition.get('driver', 'default')
        assert isinstance(driver_name, basestring)
        return driver_name

    @property
    def driver(self):
        """Return the driver for this hypervisor"""
        return self._manager.config.make_factory("virt/drivers/%s" % (self.driver_name,), self._manager)

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<Hypervisor %s at %x>" % (self.node_id, id(self))



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

import zope.interface as ZI

class IDriver(ZI.Interface):
    """
    Interface required for managing VMs on a single server.

    If you want to use the vagoth.virt.virtualmachine_ and
    vagoth.virt.hypervisor_ as-is with your hypervisor API,
    you can implement your own driver using IDriver.

    For example, you could write a driver to manage Xen, VMWare or RHEV VMs,
    or even to manage VMs in remote clouds.
    """
    def __init__(manager, config):
        """
        """
    def provision(node, vm):
        """
        Request the node to provision the VM (eg. define then re-init)
        """
    def define(node, vm):
        """
        Request the node to define the VM
        """
    def undefine(node, vm):
        """
        Request the node to undefine the VM
        """
    def deprovision(node, vm):
        """
        Request the node to deprovision the VM (eg. undefine & wipe disks)
        """
    def start(node, vm):
        """
        Request the node to start the VM (ie. power-on)
        """
    def reboot(node, vm):
        """
        Request the node to reboot the VM (poweroff/poweron)
        """
    def stop(node, vm):
        """
        Request the node to stop the VM (power-off)
        """
    def shutdown(node, vm):
        """
        Request the node to nicely shutdown the VM (eg. ACPI poweroff)
        """
    def info(node, vm):
        """
        Request the information dict about this VM.
        It should include "definition" with a dict, as well as "state",
        and possibly others.
        """
    def status(node=None):
        """
        Request the "status" for the entire node. This should include a
        "vms" dict, containing the info for each VM.
        """
    def migrate(node, vm, destination_node):
        """
        Request the node to migrate the VM to the destination_node.
        """



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

"""
Actions are called by the job scheduler to perform tasks.
They contain a dictionary of arguments, but will be called
with an instance of Vagoth's Manager as the first argument.
"""

from ..exceptions import ActionException
from .. import get_manager
from .. import transaction
import logging

Manager = get_manager()
Allocator = Manager.config.make_factory("virt/allocator", context=Manager)

def log_vm_action(vm_name, action, msg=None):
    """
    Call logging.info(), using transaction.get_txid() and .get_source()
    along with the VM name and the message, if any.
    """
    txid = transaction.get_txid()
    source = transaction.get_source()
    if msg:
        logging.info("vagoth: txid={0} source={1} action={2} vm={3}: {4}".format(txid, source, action, vm_name, msg))
    else:
        logging.info("vagoth: txid={0} source={1} action={2} vm={3}".format(txid, source, action, vm_name))

def vm_define(manager, vm_name, hint=None, **kwargs):
    """
    If a VM isn't allocated to a hypervisor node, it will
    allocate it to a hypervisor node using the `Allocator`,
    and call driver.define(hypervisor, vm)
    """
    global Allocator
    log_vm_action(vm_name, "define")
    vm = manager.get_node(vm_name)
    node = vm.parent
    if node:
        return
    Allocator.allocate(vm, hint)
    vm.refresh()
    node = vm.parent
    if node:
        vm.state = "defined"
        node.driver.define(node, vm)

def vm_provision(manager, vm_name, hint=None, **kwargs):
    """
    If a VM isn't allocated to a hypervisor node, it will
    allocate it to a hypervisor node using the `Allocator`,
    and call driver.provision(hypervisor, vm)
    """
    global Allocator
    log_vm_action(vm_name, "provision")
    vm = manager.get_node(vm_name)
    node = vm.parent
    if node:
        return
    Allocator.allocate(vm, hint)
    vm.refresh()
    node = vm.parent
    if node:
        vm.state = "defined"
        node.driver.define(node, vm)

def vm_start(manager, vm_name, **kwargs):
    """
    If a VM isn't allocated to a hypervisor node,
    it will call the vm_define action, and if that
    worked, call driver.start(hypervisor, vm)
    """
    log_vm_action(vm_name, "start")
    vm = manager.get_node(vm_name)
    node = vm.parent
    if not node:
        manager.action("vm_define", vm_name=vm_name, **kwargs)
        vm.refresh() # pick up state change
        node = vm.parent
    if node:
        vm.state = "starting"
        node.driver.start(node, vm)
    else:
        raise ActionException("VM not assigned")

def vm_stop(manager, vm_name, **kwargs):
    """
    call driver.stop(hypervisor, vm)
    """
    log_vm_action(vm_name, "stop")
    vm = manager.get_node(vm_name)
    node = vm.parent
    if node:
        vm.state = "stopping"
        node.driver.stop(node, vm)

def vm_shutdown(manager, vm_name, **kwargs):
    """
    call driver.stop(hypervisor, vm)
    """
    log_vm_action(vm_name, "stop")
    vm = manager.get_node(vm_name)
    node = vm.parent
    if node:
        vm.state = "shutting down"
        node.driver.shutdown(node, vm)

def vm_reboot(manager, vm_name, **kwargs):
    """
    call driver.reboot(hypervisor, vm)
    """
    log_vm_action(vm_name, "reboot")
    vm = manager.get_node(vm_name)
    node = vm.parent
    if node:
        vm.state = "rebooting"
        node.driver.reboot(node, vm)

def vm_undefine(manager, vm_name):
    """
    call driver.undefine(hypervisor, vm)
    """
    log_vm_action(vm_name, "undefine")
    vm = manager.get_node(vm_name)
    node = vm.parent
    if not node:
        return
    node.driver.undefine(node, vm)

def vm_deprovision(manager, vm_name):
    """
    call driver.deprovision(hypervisor, vm)
    """
    log_vm_action(vm_name, "deprovision")
    vm = manager.get_node(vm_name)
    node = vm.parent
    if not node:
        return
    node.driver.deprovision(node, vm)

def vm_poll(manager, **kwargs):
    """
    instantiate the monitor and poll all nodes
    """
    monitor = Manager.config.make_factory("virt/monitor", context=Manager)
    monitor.poll_nodes()



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

from ..utils.mc_json_rpc import mcollective_call
from ..exceptions import DriverException

ERRMSG = "Cannot manage VMs on a Hypervisor using DummyDriver"

class DummyDriver(object):
    """
    The Dummy Driver never has info and raises a DriverException if you try to use it.
    """
    def __init__(self, manager, local_config):
        self.config = local_config

    def provision(self, node, vm):
        """Request a node to define & provision a VM"""
        raise DriverException(ERRMSG)

    def define(self, node, vm):
        """Request node to define a VM"""
        raise DriverException(ERRMSG)

    def undefine(self, node, vm):
        """Request node to undefine a VM"""
        raise DriverException(ERRMSG)

    def deprovision(self, node, vm):
        """Request node to undefine & deprovision a VM"""
        raise DriverException(ERRMSG)

    def start(self, node, vm):
        """Request node to start the VM"""
        raise DriverException(ERRMSG)

    def reboot(self, node, vm):
        """Request node to reboot the VM"""
        raise DriverException(ERRMSG)

    def stop(self, node, vm):
        """Request node to stop (forcefully) the VM"""
        raise DriverException(ERRMSG)

    def shutdown(self, node, vm):
        """Request node to shutdown (nicely) the VM"""
        raise DriverException(ERRMSG)

    def info(self, node, vm):
        """Request information about the given VM from the node"""
        return {}

    def status(self, node=None):
        """Request information about all VMs from the node"""
        return []

    def migrate(self, node, vm, destination_node):
        """Request the node to migrate the given VM to the destination node"""
        return NotImplemented



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

class DummyAllocator(object):
    def __init__(self, manager, allocator_config):
        self.config = allocator_config
        self.registry = manager.registry
    def _find_best_host_for(self, vm, hint=None):
        if hint:
            return hint
        return self.config.get('node', None)
    def allocate(self, vm, hint):
        hv_node_id = self._find_best_host_for(vm, hint)
        if hv_node_id:
            self.registry.set_parent(vm.node_id, hv_node_id)
            return hv_node_id



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

def get_config(config_searchpaths):
    """
    Given a list of config searchpaths, instantiate the config object for
    the first path, or instantiate an empty config object.  This uses
    the configobj library to load the config.
    """
    from configobj import ConfigObj
    import os.path

    for path in config_searchpaths:
        path = os.path.expanduser(path)
        if os.path.exists(path):
            return ConfigObj(path)
    return ConfigObj()

def dynamic_lookup(moduleColonName):
    """
    Dynamically lookup an object in a module, given input of the form
    modulename:objectname
    """
    modulestr, name = moduleColonName.split(":")
    try:
        module = __import__(modulestr, fromlist=[name])
    except ImportError:
        raise ImportError("Could not import module: %s" % (modulestr,))
    try:
        return getattr(module, name)
    except AttributeError:
        raise AttributeError("Could not load %s from module %s" % (name, modulestr))

_static_config = None

class Config(object):
    """
    Config is a wrapper class around the real config object,
    supplying methods to return or instantate class factories
    found in the configuration.
    """
    def __init__(self, config_searchpaths=None):
        global _static_config
        if _static_config != None:
            self.config = _static_config
        else:
            searchpath = config_searchpaths or ["~/.config/sippe/vagoth.conf", "/etc/sippe/vagoth.conf"]
            _static_config = self.config = get_config(searchpath)

    def __getitem__(self, key):
        return self.config[key]

    def _lookup_config_section(self, config_path):
        current = self.config
        try:
            for part in config_path.split("/"):
                current = current[part]
            return current
        except KeyError:
            raise KeyError("Could not locate %s in config file" % (config_path,))

    def get_factory(self, section, factory_key="factory"):
        """Return a tuple for a factory and its configuration dict"""
        config = self._lookup_config_section(section)
        factory_str = config[factory_key]
        return dynamic_lookup(factory_str), config

    def make_factory(self, section, context, factory_key="factory"):
        """
        Instantate a factory, passing in a context object and the section
        config to the initialiser.
        """
        factory_class, factory_config = self.get_factory(section, factory_key)
        return factory_class(context, factory_config)

    def get_node_factory(self, node_type):
        """
        Return the class for a given node type
        """
        node_types = self.config["node_types"]
        if node_type in node_types:
            return dynamic_lookup(node_types[node_type])

    def get_action(self, action):
        """
        Return the named action callable
        """
        actions = self.config["actions"]
        if action in actions:
            return dynamic_lookup(actions[action])



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

"""
Wrap any code block in a transaction in order to get
a transaction id in the logs.

Example:

>>> with Transaction(username):
...    manager.action("start", vm_name="my_vm")

Nested transaction's will retain the first transaction ID, instead of generating a
new one.
"""

import threading
import uuid

threadlocal = threading.local()
threadlocal.vagoth_txid = None
threadlocal.source = None

def get_txid():
    """
    Return the current transaction ID, or "0"
    """
    global threadlocal
    try:
        return threadlocal.vagoth_txid or "0"
    except AttributeError:
        return "0"

def get_source():
    """
    Return the current transaction source, or "0"
    """
    global threadlocal
    try:
        return threadlocal.source or "0"
    except AttributeError:
        return "0"

class Transaction(object):
    """
    Transaction(source=None, txid=None)

    source might be a username, or the name of the process.

    If a transaction is already set for the current thread, it
    will be used instead of the passed-in txid.

    If txid is not specified, a random one will be generated.
    """
    def __init__(self, source=None, txid=None):
        self.source = source
        self.desired_txid = txid

    def __enter__(self):
        global threadlocal
        self.txid = None
        if threadlocal.vagoth_txid == None:
            self.txid = self.desired_txid or uuid.uuid4().hex[:8]
            threadlocal.vagoth_txid = self.txid
            threadlocal.source = self.source

    def __exit__(self, *args, **kwargs):
        global threadlocal
        if self.txid != None:
            threadlocal.vagoth_txid = None
            threadlocal.source = None



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

import zope.interface as ZI

class IScheduler(ZI.Interface):
    """
    Scheduler's are called to schedule the background execution of actions.
    Vagoth's default schedulers are quite basic, but you can implement your
    own so long as it matches the IScheduler interface.
    """
    def __init__(manager, config):
        """
        Instantiated with an instance of Manager and a configuration dict
        """

    def action(self, queue_name, action, **kwargs):
        """
        Schedule the given action with kwargs.

        Actions in the same queue_name should be executed in sequence.
        """

    def cleanup(self):
        """
        Called by Manager.cleanup() at shutdown time.
        It could be used to close DB connections, cleanup threads or
        child processes, etc.
        """



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

#
# A wrapper for a Node Document/Dict
#

import zope.interface as ZI

class INodeDoc(ZI.Interface):
    """
    INodeDoc represents a node as returned from an IRegistry.
    """
    type = ZI.Attribute("node type as string")
    id = ZI.Attribute("node id as string")
    name = ZI.Attribute("node name as string")
    tenant = ZI.Attribute("node tenant as string")
    definition = ZI.Attribute("node definition as dict")
    metadata = ZI.Attribute("node metadata as dict")
    parent = ZI.Attribute("node parent")
    tags = ZI.Attribute("node tags as dict")

    def get_blob(key):
        """Return the node blob for the given key (or None)"""



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

#
# A wrapper for a Node Document/Dict
#

class NodeDoc(object):
    """
    An encapsulation of the node dictionary, implementing INodeDoc
    """
    def __init__(self, registry, node_dict):
        self.registry = registry
        self.node_dict = node_dict
        assert "type" in node_dict
        assert "node_id" in node_dict
        assert "name" in node_dict
        assert "parent" in node_dict
        assert "definition" in node_dict
        assert "unique_keys" in node_dict
        if "tenant" not in node_dict:
            node_dict["tenant"] = None
        if "metadata" not in node_dict:
            node_dict["metadata"] = {}
        if "tags" not in node_dict:
            node_dict["tags"] = {}
        elif type(node_dict["tags"]) == list: # migrate to dict..
            node_dict["tags"] = dict([(x,True) for x in node_dict["tags"]])

    @property
    def type(self):
        return self.node_dict["type"]

    @property
    def id(self):
        return self.node_dict["node_id"]

    @property
    def name(self):
        return self.node_dict["name"]

    @property
    def definition(self):
        return self.node_dict["definition"]

    @property
    def metadata(self):
        return self.node_dict["metadata"]

    @property
    def parent(self):
        return self.node_dict["parent"]

    @property
    def tags(self):
        return self.node_dict["tags"]

    @property
    def unique_keys(self):
        return self.node_dict["unique_keys"]

    @property
    def tenant(self):
        return self.node_dict["tenant"]

    def get_blob(self, key):
        return self.registry.get_blob(self.id, key)

    def __getitem__(self, key):
        raise NotImplementedError

    def get(self, *args, **kwargs):
        raise NotImplementedError

    def __str__(self):
        return "NodeDoc for %s" % (self.id,)

    def __repr__(self):
        return "<NodeDoc for %s at 0x%x>" % (self.id, id(self))



#
# Vagoth Cluster Management Framework
# Copyright (C) 2013  Robert Thomson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#

from ..node import Node

class VirtualMachine(Node):
    """
    `VirtualMachine` represents a VirtualMachine node type.

    It inherits vagoth.Node_

    Each method requests the scheduler to execute the related action.
    """

    @property
    def state(self):
        """Retrieve the state attribute from the metadata"""
        return self._doc.metadata.get('state', 'unknown')

    @state.setter
    def state(self, state):
        """Set the state attribute in the metadata"""
        self._doc.registry.update_metadata(self.node_id, { "state": state, })
        self.refresh()

    def start(self, hint=None):
        """Schedule the start action"""
        self._manager.scheduler.action(self.node_id, "vm_start", vm_name=self.node_id, hint=hint)

    def define(self, hint=None):
        """Schedule the define action"""
        self._manager.scheduler.action(self.node_id, "vm_define", vm_name=self.node_id, hint=hint)

    def stop(self):
        """Schedule the stop action"""
        self._manager.scheduler.action(self.node_id, "vm_stop", vm_name=self.node_id)

    def shutdown(self):
        """Schedule the shutdown action"""
        self._manager.scheduler.action(self.node_id, "vm_shutdown", vm_name=self.node_id)

    def reboot(self):
        """Schedule the reboot action"""
        self._manager.scheduler.action(self.node_id, "vm_reboot", vm_name=self.node_id)

    def undefine(self):
        """Schedule the undefine action"""
        self._manager.scheduler.action(self.node_id, "vm_undefine", vm_name=self.node_id)

    def provision(self):
        """Schedule the provision action"""
        self._manager.scheduler.action(self.node_id, "vm_provision", vm_name=self.node_id)

    def deprovision(self):
        """Schedule the deprovision action"""
        self._manager.scheduler.action(self.node_id, "vm_deprovision", vm_name=self.node_id)

    def __str__(self):
        return self.node_id

    def __repr__(self):
        return "<VirtualMachine %s at %x>" % (self.node_id, id(self))

